@@ -1,27 +1,32 @@
-#include "mupdf/html.h"
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
 
-static const char *inherit_list[] = {
-	"color",
-	"direction",
-	"font-family",
-	"font-style",
-	"font-variant",
-	"font-weight",
-	"letter-spacing",
-	"line-height",
-	"list-style-image",
-	"list-style-position",
-	"list-style-type",
-	"orphans",
-	"quotes",
-	"text-align",
-	"text-indent",
-	"text-transform",
-	"visibility",
-	"white-space",
-	"widows",
-	"word-spacing",
-};
+#include "mupdf/fitz.h"
+#include "html-imp.h"
+
+#include <string.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <assert.h>
 
 static const char *border_width_kw[] = {
 	"medium",
@@ -237,15 +242,15 @@ match_att_has_condition(fz_xml *node, const char *att, const char *needle)
 	const char *ss;
 	size_t n;
 	if (haystack) {
-		/* Try matching whole property first. */
-		if (!strcmp(haystack, needle))
-			return 1;
-
-		/* Look for matching words. */
-		n = strlen(needle);
 		ss = strstr(haystack, needle);
-		if (ss && (ss[n] == ' ' || ss[n] == 0) && (ss == haystack || ss[-1] == ' '))
-			return 1;
+		if (ss)
+		{
+			n = strlen(needle);
+
+			/* Look for exact matches or matching words. */
+			if ((ss[n] == ' ' || ss[n] == 0) && (ss == haystack || ss[-1] == ' '))
+				return 1;
+		}
 	}
 	return 0;
 }
@@ -282,11 +287,13 @@ match_selector(fz_css_selector *sel, fz_xml *node)
 		if (sel->combine == ' ')
 		{
 			fz_xml *parent = fz_xml_up(node);
+			if (!parent || !match_selector(sel->right, node))
+				return 0;
+
 			while (parent)
 			{
 				if (match_selector(sel->left, parent))
-					if (match_selector(sel->right, node))
-						return 1;
+					return 1;
 				parent = fz_xml_up(parent);
 			}
 			return 0;
@@ -323,7 +330,7 @@ match_selector(fz_css_selector *sel, fz_xml *node)
 
 	if (sel->name)
 	{
-		if (strcmp(sel->name, fz_xml_tag(node)))
+		if (!fz_xml_is_tag(node, sel->name))
 			return 0;
 	}
 
@@ -352,11 +359,11 @@ count_values(fz_css_value *value)
 	return n;
 }
 
-static void add_property(fz_css_match *match, const char *name, fz_css_value *value, int spec);
+static void add_property(fz_css_match *match, int name, fz_css_value *value, int spec);
 
 static void
 add_shorthand_trbl(fz_css_match *match, fz_css_value *value, int spec,
-	const char *name_t, const char *name_r, const char *name_b, const char *name_l)
+	int name_t, int name_r, int name_b, int name_l)
 {
 	int n = count_values(value);
 
@@ -409,35 +416,50 @@ static void
 add_shorthand_margin(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"margin-top", "margin-right", "margin-bottom", "margin-left");
+		PRO_MARGIN_TOP,
+		PRO_MARGIN_RIGHT,
+		PRO_MARGIN_BOTTOM,
+		PRO_MARGIN_LEFT);
 }
 
 static void
 add_shorthand_padding(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"padding-top", "padding-right", "padding-bottom", "padding-left");
+		PRO_PADDING_TOP,
+		PRO_PADDING_RIGHT,
+		PRO_PADDING_BOTTOM,
+		PRO_PADDING_LEFT);
 }
 
 static void
 add_shorthand_border_width(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"border-top-width", "border-right-width", "border-bottom-width", "border-left-width");
+		PRO_BORDER_TOP_WIDTH,
+		PRO_BORDER_RIGHT_WIDTH,
+		PRO_BORDER_BOTTOM_WIDTH,
+		PRO_BORDER_LEFT_WIDTH);
 }
 
 static void
 add_shorthand_border_color(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"border-top-color", "border-right-color", "border-bottom-color", "border-left-color");
+		PRO_BORDER_TOP_COLOR,
+		PRO_BORDER_RIGHT_COLOR,
+		PRO_BORDER_BOTTOM_COLOR,
+		PRO_BORDER_LEFT_COLOR);
 }
 
 static void
 add_shorthand_border_style(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"border-top-style", "border-right-style", "border-bottom-style", "border-left-style");
+		PRO_BORDER_TOP_STYLE,
+		PRO_BORDER_RIGHT_STYLE,
+		PRO_BORDER_BOTTOM_STYLE,
+		PRO_BORDER_LEFT_STYLE);
 }
 
 static void
@@ -447,41 +469,41 @@ add_shorthand_border(fz_css_match *match, fz_css_value *value, int spec, int T,
 	{
 		if (value->type == CSS_HASH)
 		{
-			if (T) add_property(match, "border-top-color", value, spec);
-			if (R) add_property(match, "border-right-color", value, spec);
-			if (B) add_property(match, "border-bottom-color", value, spec);
-			if (L) add_property(match, "border-left-color", value, spec);
+			if (T) add_property(match, PRO_BORDER_TOP_COLOR, value, spec);
+			if (R) add_property(match, PRO_BORDER_RIGHT_COLOR, value, spec);
+			if (B) add_property(match, PRO_BORDER_BOTTOM_COLOR, value, spec);
+			if (L) add_property(match, PRO_BORDER_LEFT_COLOR, value, spec);
 		}
 		else if (value->type == CSS_KEYWORD)
 		{
 			if (keyword_in_list(value->data, border_width_kw, nelem(border_width_kw)))
 			{
-				if (T) add_property(match, "border-top-width", value, spec);
-				if (R) add_property(match, "border-right-width", value, spec);
-				if (B) add_property(match, "border-bottom-width", value, spec);
-				if (L) add_property(match, "border-left-width", value, spec);
+				if (T) add_property(match, PRO_BORDER_TOP_WIDTH, value, spec);
+				if (R) add_property(match, PRO_BORDER_RIGHT_WIDTH, value, spec);
+				if (B) add_property(match, PRO_BORDER_BOTTOM_WIDTH, value, spec);
+				if (L) add_property(match, PRO_BORDER_LEFT_WIDTH, value, spec);
 			}
 			else if (keyword_in_list(value->data, border_style_kw, nelem(border_style_kw)))
 			{
-				if (T) add_property(match, "border-top-style", value, spec);
-				if (R) add_property(match, "border-right-style", value, spec);
-				if (B) add_property(match, "border-bottom-style", value, spec);
-				if (L) add_property(match, "border-left-style", value, spec);
+				if (T) add_property(match, PRO_BORDER_TOP_STYLE, value, spec);
+				if (R) add_property(match, PRO_BORDER_RIGHT_STYLE, value, spec);
+				if (B) add_property(match, PRO_BORDER_BOTTOM_STYLE, value, spec);
+				if (L) add_property(match, PRO_BORDER_LEFT_STYLE, value, spec);
 			}
 			else if (keyword_in_list(value->data, color_kw, nelem(color_kw)))
 			{
-				if (T) add_property(match, "border-top-color", value, spec);
-				if (R) add_property(match, "border-right-color", value, spec);
-				if (B) add_property(match, "border-bottom-color", value, spec);
-				if (L) add_property(match, "border-left-color", value, spec);
+				if (T) add_property(match, PRO_BORDER_TOP_COLOR, value, spec);
+				if (R) add_property(match, PRO_BORDER_RIGHT_COLOR, value, spec);
+				if (B) add_property(match, PRO_BORDER_BOTTOM_COLOR, value, spec);
+				if (L) add_property(match, PRO_BORDER_LEFT_COLOR, value, spec);
 			}
 		}
 		else
 		{
-			if (T) add_property(match, "border-top-width", value, spec);
-			if (R) add_property(match, "border-right-width", value, spec);
-			if (B) add_property(match, "border-bottom-width", value, spec);
-			if (L) add_property(match, "border-left-width", value, spec);
+			if (T) add_property(match, PRO_BORDER_TOP_WIDTH, value, spec);
+			if (R) add_property(match, PRO_BORDER_RIGHT_WIDTH, value, spec);
+			if (B) add_property(match, PRO_BORDER_BOTTOM_WIDTH, value, spec);
+			if (L) add_property(match, PRO_BORDER_LEFT_WIDTH, value, spec);
 		}
 		value = value->next;
 	}
@@ -496,11 +518,11 @@ add_shorthand_list_style(fz_css_match *match, fz_css_value *value, int spec)
 		{
 			if (keyword_in_list(value->data, list_style_type_kw, nelem(list_style_type_kw)))
 			{
-				add_property(match, "list-style-type", value, spec);
+				add_property(match, PRO_LIST_STYLE_TYPE, value, spec);
 			}
 			else if (keyword_in_list(value->data, list_style_position_kw, nelem(list_style_position_kw)))
 			{
-				add_property(match, "list-style-position", value, spec);
+				add_property(match, PRO_LIST_STYLE_POSITION, value, spec);
 			}
 		}
 		value = value->next;
@@ -508,126 +530,70 @@ add_shorthand_list_style(fz_css_match *match, fz_css_value *value, int spec)
 }
 
 static void
-add_property(fz_css_match *match, const char *name, fz_css_value *value, int spec)
+add_property(fz_css_match *match, int name, fz_css_value *value, int spec)
 {
-	int i;
-
-	if (!strcmp(name, "margin"))
+	/* shorthand expansions: */
+	switch (name)
 	{
+	case PRO_MARGIN:
 		add_shorthand_margin(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "padding"))
-	{
+	case PRO_PADDING:
 		add_shorthand_padding(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "border-width"))
-	{
+	case PRO_BORDER_WIDTH:
 		add_shorthand_border_width(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "border-color"))
-	{
+	case PRO_BORDER_COLOR:
 		add_shorthand_border_color(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "border-style"))
-	{
+	case PRO_BORDER_STYLE:
 		add_shorthand_border_style(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "border"))
-	{
+	case PRO_BORDER:
 		add_shorthand_border(match, value, spec, 1, 1, 1, 1);
 		return;
-	}
-	if (!strcmp(name, "border-top"))
-	{
+	case PRO_BORDER_TOP:
 		add_shorthand_border(match, value, spec, 1, 0, 0, 0);
 		return;
-	}
-	if (!strcmp(name, "border-right"))
-	{
+	case PRO_BORDER_RIGHT:
 		add_shorthand_border(match, value, spec, 0, 1, 0, 0);
 		return;
-	}
-	if (!strcmp(name, "border-bottom"))
-	{
+	case PRO_BORDER_BOTTOM:
 		add_shorthand_border(match, value, spec, 0, 0, 1, 0);
 		return;
-	}
-	if (!strcmp(name, "border-left"))
-	{
+	case PRO_BORDER_LEFT:
 		add_shorthand_border(match, value, spec, 0, 0, 0, 1);
 		return;
-	}
-	if (!strcmp(name, "list-style"))
-	{
+	case PRO_LIST_STYLE:
 		add_shorthand_list_style(match, value, spec);
 		return;
-	}
-
-	/* shorthand expansions: */
-	/* TODO: border-color */
-	/* TODO: border-style */
 	/* TODO: font */
-	/* TODO: list-style */
 	/* TODO: background */
-
-	for (i = 0; i < match->count; ++i)
-	{
-		if (!strcmp(match->prop[i].name, name))
-		{
-			if (match->prop[i].spec <= spec)
-			{
-				match->prop[i].value = value;
-				match->prop[i].spec = spec;
-			}
-			return;
-		}
 	}
 
-	if (match->count + 1 >= nelem(match->prop))
+	if (name < NUM_PROPERTIES && match->spec[name] <= spec)
 	{
-		// fz_warn(ctx, "too many css properties");
-		return;
-	}
-
-	match->prop[match->count].name = name;
-	match->prop[match->count].value = value;
-	match->prop[match->count].spec = spec;
-	++match->count;
-}
-
-static void
-sort_properties(fz_css_match *match)
-{
-	int count = match->count;
-	fz_css_match_prop *prop = match->prop;
-	int i, k;
-
-	/* Insertion sort. */
-	for (i = 1; i < count; ++i)
-	{
-		k = i;
-		while (k > 0 && strcmp(prop[k-1].name, prop[k].name) > 0)
-		{
-			fz_css_match_prop save = prop[k-1];
-			prop[k-1] = prop[k];
-			prop[k] = save;
-			--k;
-		}
+		match->value[name] = value;
+		match->spec[name] = spec;
 	}
 }
 
 void
-fz_match_css(fz_context *ctx, fz_css_match *match, fz_css *css, fz_xml *node)
+fz_match_css(fz_context *ctx, fz_css_match *match, fz_css_match *up, fz_css *css, fz_xml *node)
 {
 	fz_css_rule *rule;
 	fz_css_selector *sel;
 	fz_css_property *prop;
 	const char *s;
+	int i;
+
+	match->up = up;
+	for (i = 0; i < NUM_PROPERTIES; ++i)
+	{
+		match->spec[i] = -1;
+		match->value[i] = NULL;
+	}
 
 	for (rule = css->rule; rule; rule = rule->next)
 	{
@@ -665,8 +631,6 @@ fz_match_css(fz_context *ctx, fz_css_match *match, fz_css *css, fz_xml *node)
 			}
 		}
 	}
-
-	sort_properties(match); /* speed up subsequent value_from_raw_property lookups */
 }
 
 void
@@ -675,6 +639,14 @@ fz_match_css_at_page(fz_context *ctx, fz_css_match *match, fz_css *css)
 	fz_css_rule *rule;
 	fz_css_selector *sel;
 	fz_css_property *prop;
+	int i;
+
+	match->up = NULL;
+	for (i = 0; i < NUM_PROPERTIES; ++i)
+	{
+		match->spec[i] = -1;
+		match->value[i] = NULL;
+	}
 
 	for (rule = css->rule; rule; rule = rule->next)
 	{
@@ -690,8 +662,6 @@ fz_match_css_at_page(fz_context *ctx, fz_css_match *match, fz_css *css)
 			sel = sel->next;
 		}
 	}
-
-	sort_properties(match); /* speed up subsequent value_from_raw_property lookups */
 }
 
 void
@@ -701,20 +671,22 @@ fz_add_css_font_face(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, co
 	fz_css_property *prop;
 	fz_font *font = NULL;
 	fz_buffer *buf = NULL;
-	int is_bold, is_italic;
+	int is_bold, is_italic, is_small_caps;
 	char path[2048];
 
 	const char *family = "serif";
 	const char *weight = "normal";
 	const char *style = "normal";
+	const char *variant = "normal";
 	const char *src = NULL;
 
 	for (prop = declaration; prop; prop = prop->next)
 	{
-		if (!strcmp(prop->name, "font-family")) family = prop->value->data;
-		if (!strcmp(prop->name, "font-weight")) weight = prop->value->data;
-		if (!strcmp(prop->name, "font-style")) style = prop->value->data;
-		if (!strcmp(prop->name, "src")) src = prop->value->data;
+		if (prop->name == PRO_FONT_FAMILY) family = prop->value->data;
+		if (prop->name == PRO_FONT_WEIGHT) weight = prop->value->data;
+		if (prop->name == PRO_FONT_STYLE) style = prop->value->data;
+		if (prop->name == PRO_FONT_VARIANT) variant = prop->value->data;
+		if (prop->name == PRO_SRC) src = prop->value->data;
 	}
 
 	if (!src)
@@ -722,6 +694,7 @@ fz_add_css_font_face(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, co
 
 	is_bold = is_bold_from_font_weight(weight);
 	is_italic = is_italic_from_font_style(style);
+	is_small_caps = !strcmp(variant, "small-caps");
 
 	fz_strlcpy(path, base_uri, sizeof path);
 	fz_strlcat(path, "/", sizeof path);
@@ -732,11 +705,10 @@ fz_add_css_font_face(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, co
 	for (custom = set->custom; custom; custom = custom->next)
 		if (!strcmp(custom->src, path) && !strcmp(custom->family, family) &&
 				custom->is_bold == is_bold &&
-				custom->is_italic == is_italic)
+				custom->is_italic == is_italic &&
+				custom->is_small_caps == is_small_caps)
 			return; /* already loaded */
 
-	printf("epub: @font-face: family='%s' b=%d i=%d src=%s\n", family, is_bold, is_italic, src);
-
 	fz_var(buf);
 	fz_var(font);
 
@@ -746,8 +718,8 @@ fz_add_css_font_face(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, co
 			buf = fz_read_archive_entry(ctx, zip, path);
 		else
 			buf = fz_read_file(ctx, src);
-		font = fz_new_font_from_buffer(ctx, src, buf, 0, 0);
-		fz_add_html_font_face(ctx, set, family, is_bold, is_italic, path, font);
+		font = fz_new_font_from_buffer(ctx, NULL, buf, 0, 0);
+		fz_add_html_font_face(ctx, set, family, is_bold, is_italic, is_small_caps, path, font);
 	}
 	fz_always(ctx)
 	{
@@ -756,6 +728,7 @@ fz_add_css_font_face(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, co
 	}
 	fz_catch(ctx)
 	{
+		fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 		fz_warn(ctx, "cannot load font-face: %s", src);
 	}
 }
@@ -768,61 +741,82 @@ fz_add_css_font_faces(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, c
 
 	for (rule = css->rule; rule; rule = rule->next)
 	{
-		sel = rule->selector;
-		while (sel)
+		if (!rule->loaded)
 		{
-			if (sel->name && !strcmp(sel->name, "@font-face"))
+			rule->loaded = 1;
+			sel = rule->selector;
+			while (sel)
 			{
-				fz_add_css_font_face(ctx, set, zip, base_uri, rule->declaration);
-				break;
+				if (sel->name && !strcmp(sel->name, "@font-face"))
+				{
+					fz_add_css_font_face(ctx, set, zip, base_uri, rule->declaration);
+					break;
+				}
+				sel = sel->next;
 			}
-			sel = sel->next;
 		}
 	}
 }
 
-static fz_css_value *
-value_from_raw_property(fz_css_match *match, const char *name)
+static int
+is_inheritable_property(int name)
 {
-	fz_css_match_prop *prop = match->prop;
-	int l = 0;
-	int r = match->count - 1;
-	while (l <= r)
-	{
-		int m = (l + r) >> 1;
-		int c = strcmp(name, prop[m].name);
-		if (c < 0)
-			r = m - 1;
-		else if (c > 0)
-			l = m + 1;
-		else
-			return prop[m].value;
-	}
-	return NULL;
+	return
+		name == PRO_COLOR ||
+		name == PRO_DIRECTION ||
+		name == PRO_FONT_FAMILY ||
+		name == PRO_FONT_STYLE ||
+		name == PRO_FONT_VARIANT ||
+		name == PRO_FONT_WEIGHT ||
+		name == PRO_LETTER_SPACING ||
+		name == PRO_LINE_HEIGHT ||
+		name == PRO_LIST_STYLE_IMAGE ||
+		name == PRO_LIST_STYLE_POSITION ||
+		name == PRO_LIST_STYLE_TYPE ||
+		name == PRO_ORPHANS ||
+		name == PRO_QUOTES ||
+		name == PRO_TEXT_ALIGN ||
+		name == PRO_TEXT_INDENT ||
+		name == PRO_TEXT_TRANSFORM ||
+		name == PRO_VISIBILITY ||
+		name == PRO_WHITE_SPACE ||
+		name == PRO_WIDOWS ||
+		name == PRO_WORD_SPACING;
 }
 
 static fz_css_value *
-value_from_property(fz_css_match *match, const char *name)
+value_from_inheritable_property(fz_css_match *match, int name)
 {
-	fz_css_value *value;
-
-	value = value_from_raw_property(match, name);
+	fz_css_value *value = match->value[name];
 	if (match->up)
 	{
 		if (value && !strcmp(value->data, "inherit"))
-			if (strcmp(name, "font-size") != 0) /* never inherit 'font-size' textually */
+			return value_from_inheritable_property(match->up, name);
+		if (!value)
+			return value_from_inheritable_property(match->up, name);
+	}
+	return value;
+}
+
+static fz_css_value *
+value_from_property(fz_css_match *match, int name)
+{
+	fz_css_value *value = match->value[name];
+	if (match->up)
+	{
+		if (value && !strcmp(value->data, "inherit"))
+			if (name != PRO_FONT_SIZE) /* never inherit 'font-size' textually */
 				return value_from_property(match->up, name);
-		if (!value && keyword_in_list(name, inherit_list, nelem(inherit_list)))
-			return value_from_property(match->up, name);
+		if (!value && is_inheritable_property(name))
+			return value_from_inheritable_property(match->up, name);
 	}
 	return value;
 }
 
 static const char *
-string_from_property(fz_css_match *match, const char *name, const char *initial)
+string_from_property(fz_css_match *match, int name, const char *initial)
 {
-	fz_css_value *value;
-	value = value_from_property(match, name);
+	fz_css_value *value = value_from_property(match, name);
 	if (!value)
 		return initial;
 	return value->data;
@@ -934,13 +928,13 @@ number_from_value(fz_css_value *value, float initial, int initial_unit)
 }
 
 static fz_css_number
-number_from_property(fz_css_match *match, const char *property, float initial, int initial_unit)
+number_from_property(fz_css_match *match, int property, float initial, int initial_unit)
 {
 	return number_from_value(value_from_property(match, property), initial, initial_unit);
 }
 
 static fz_css_number
-border_width_from_property(fz_css_match *match, const char *property)
+border_width_from_property(fz_css_match *match, int property)
 {
 	fz_css_value *value = value_from_property(match, property);
 	if (value)
@@ -957,7 +951,7 @@ border_width_from_property(fz_css_match *match, const char *property)
 }
 
 static int
-border_style_from_property(fz_css_match *match, const char *property)
+border_style_from_property(fz_css_match *match, int property)
 {
 	fz_css_value *value = value_from_property(match, property);
 	if (value)
@@ -970,28 +964,28 @@ border_style_from_property(fz_css_match *match, const char *property)
 }
 
 float
-fz_from_css_number(fz_css_number number, float em, float width)
+fz_from_css_number(fz_css_number number, float em, float percent_value, float auto_value)
 {
 	switch (number.unit) {
 	default:
 	case N_NUMBER: return number.value;
 	case N_LENGTH: return number.value;
 	case N_SCALE: return number.value * em;
-	case N_PERCENT: return number.value * 0.01 * width;
-	case N_AUTO: return width;
+	case N_PERCENT: return number.value * 0.01f * percent_value;
+	case N_AUTO: return auto_value;
 	}
 }
 
 float
-fz_from_css_number_scale(fz_css_number number, float scale, float em, float width)
+fz_from_css_number_scale(fz_css_number number, float scale)
 {
 	switch (number.unit) {
 	default:
 	case N_NUMBER: return number.value * scale;
 	case N_LENGTH: return number.value;
-	case N_SCALE: return number.value * em;
-	case N_PERCENT: return number.value * 0.01 * width;
-	case N_AUTO: return width;
+	case N_SCALE: return number.value * scale;
+	case N_PERCENT: return number.value * 0.01f * scale;
+	case N_AUTO: return scale;
 	}
 }
 
@@ -1021,7 +1015,7 @@ color_from_value(fz_css_value *value, fz_css_color initial)
 
 	if (value->type == CSS_HASH)
 	{
-		int r, g, b;
+		int r, g, b, a;
 		size_t n;
 hex_color:
 		n = strlen(value->data);
@@ -1030,18 +1024,35 @@ hex_color:
 			r = tohex(value->data[0]) * 16 + tohex(value->data[0]);
 			g = tohex(value->data[1]) * 16 + tohex(value->data[1]);
 			b = tohex(value->data[2]) * 16 + tohex(value->data[2]);
+			a = 255;
+		}
+		else if (n == 4)
+		{
+			r = tohex(value->data[0]) * 16 + tohex(value->data[0]);
+			g = tohex(value->data[1]) * 16 + tohex(value->data[1]);
+			b = tohex(value->data[2]) * 16 + tohex(value->data[2]);
+			a = tohex(value->data[3]) * 16 + tohex(value->data[3]);
 		}
 		else if (n == 6)
 		{
 			r = tohex(value->data[0]) * 16 + tohex(value->data[1]);
 			g = tohex(value->data[2]) * 16 + tohex(value->data[3]);
 			b = tohex(value->data[4]) * 16 + tohex(value->data[5]);
+			a = 255;
+		}
+		else if (n == 8)
+		{
+			r = tohex(value->data[0]) * 16 + tohex(value->data[1]);
+			g = tohex(value->data[2]) * 16 + tohex(value->data[3]);
+			b = tohex(value->data[4]) * 16 + tohex(value->data[5]);
+			a = tohex(value->data[6]) * 16 + tohex(value->data[7]);
 		}
 		else
 		{
 			r = g = b = 0;
+			a = 255;
 		}
-		return make_color(r, g, b, 255);
+		return make_color(r, g, b, a);
 	}
 
 	if (value->type == '(' && !strcmp(value->data, "rgb"))
@@ -1051,12 +1062,27 @@ hex_color:
 		vr = value->args;
 		vg = vr && vr->next ? vr->next->next : NULL; /* skip the ',' nodes */
 		vb = vg && vg->next ? vg->next->next : NULL; /* skip the ',' nodes */
-		r = fz_from_css_number(number_from_value(vr, 0, N_NUMBER), 255, 255);
-		g = fz_from_css_number(number_from_value(vg, 0, N_NUMBER), 255, 255);
-		b = fz_from_css_number(number_from_value(vb, 0, N_NUMBER), 255, 255);
+		r = fz_from_css_number(number_from_value(vr, 0, N_NUMBER), 255, 255, 0);
+		g = fz_from_css_number(number_from_value(vg, 0, N_NUMBER), 255, 255, 0);
+		b = fz_from_css_number(number_from_value(vb, 0, N_NUMBER), 255, 255, 0);
 		return make_color(r, g, b, 255);
 	}
 
+	if (value->type == '(' && !strcmp(value->data, "rgba"))
+	{
+		fz_css_value *vr, *vg, *vb, *va;
+		int r, g, b, a;
+		vr = value->args;
+		vg = vr && vr->next ? vr->next->next : NULL; /* skip the ',' nodes */
+		vb = vg && vg->next ? vg->next->next : NULL; /* skip the ',' nodes */
+		va = vb && vb->next ? vb->next->next : NULL; /* skip the ',' nodes */
+		r = fz_from_css_number(number_from_value(vr, 0, N_NUMBER), 255, 255, 0);
+		g = fz_from_css_number(number_from_value(vg, 0, N_NUMBER), 255, 255, 0);
+		b = fz_from_css_number(number_from_value(vb, 0, N_NUMBER), 255, 255, 0);
+		a = fz_from_css_number(number_from_value(va, 0, N_NUMBER), 255, 255, 255);
+		return make_color(r, g, b, a);
+	}
+
 	if (value->type == CSS_KEYWORD)
 	{
 		if (!strcmp(value->data, "transparent"))
@@ -1101,7 +1127,7 @@ hex_color:
 }
 
 static fz_css_color
-color_from_property(fz_css_match *match, const char *property, fz_css_color initial)
+color_from_property(fz_css_match *match, int property, fz_css_color initial)
 {
 	return color_from_value(value_from_property(match, property), initial);
 }
@@ -1109,7 +1135,7 @@ color_from_property(fz_css_match *match, const char *property, fz_css_color init
 int
 fz_get_css_match_display(fz_css_match *match)
 {
-	fz_css_value *value = value_from_property(match, "display");
+	fz_css_value *value = value_from_property(match, PRO_DISPLAY);
 	if (value)
 	{
 		if (!strcmp(value->data, "none"))
@@ -1122,6 +1148,12 @@ fz_get_css_match_display(fz_css_match *match)
 			return DIS_LIST_ITEM;
 		if (!strcmp(value->data, "inline-block"))
 			return DIS_INLINE_BLOCK;
+		if (!strcmp(value->data, "table"))
+			return DIS_TABLE;
+		if (!strcmp(value->data, "table-row"))
+			return DIS_TABLE_ROW;
+		if (!strcmp(value->data, "table-cell"))
+			return DIS_TABLE_CELL;
 	}
 	return DIS_INLINE;
 }
@@ -1129,7 +1161,7 @@ fz_get_css_match_display(fz_css_match *match)
 static int
 white_space_from_property(fz_css_match *match)
 {
-	fz_css_value *value = value_from_property(match, "white-space");
+	fz_css_value *value = value_from_property(match, PRO_WHITE_SPACE);
 	if (value)
 	{
 		if (!strcmp(value->data, "normal")) return WS_NORMAL;
@@ -1144,7 +1176,7 @@ white_space_from_property(fz_css_match *match)
 static int
 visibility_from_property(fz_css_match *match)
 {
-	fz_css_value *value = value_from_property(match, "visibility");
+	fz_css_value *value = value_from_property(match, PRO_VISIBILITY);
 	if (value)
 	{
 		if (!strcmp(value->data, "visible")) return V_VISIBLE;
@@ -1155,7 +1187,7 @@ visibility_from_property(fz_css_match *match)
 }
 
 static int
-page_break_from_property(fz_css_match *match, char *prop)
+page_break_from_property(fz_css_match *match, int prop)
 {
 	fz_css_value *value = value_from_property(match, prop);
 	if (value)
@@ -1195,10 +1227,10 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 
 	style->visibility = visibility_from_property(match);
 	style->white_space = white_space_from_property(match);
-	style->page_break_before = page_break_from_property(match, "page-break-before");
-	style->page_break_after = page_break_from_property(match, "page-break-after");
+	style->page_break_before = page_break_from_property(match, PRO_PAGE_BREAK_BEFORE);
+	style->page_break_after = page_break_from_property(match, PRO_PAGE_BREAK_AFTER);
 
-	value = value_from_property(match, "text-align");
+	value = value_from_property(match, PRO_TEXT_ALIGN);
 	if (value)
 	{
 		if (!strcmp(value->data, "left")) style->text_align = TA_LEFT;
@@ -1207,7 +1239,7 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		else if (!strcmp(value->data, "justify")) style->text_align = TA_JUSTIFY;
 	}
 
-	value = value_from_property(match, "vertical-align");
+	value = value_from_property(match, PRO_VERTICAL_ALIGN);
 	if (value)
 	{
 		if (!strcmp(value->data, "baseline")) style->vertical_align = VA_BASELINE;
@@ -1219,7 +1251,7 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		else if (!strcmp(value->data, "text-bottom")) style->vertical_align = VA_TEXT_BOTTOM;
 	}
 
-	value = value_from_property(match, "font-size");
+	value = value_from_property(match, PRO_FONT_SIZE);
 	if (value)
 	{
 		if (!strcmp(value->data, "xx-large")) style->font_size = make_number(1.73f, N_SCALE);
@@ -1238,7 +1270,7 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		style->font_size = make_number(1, N_SCALE);
 	}
 
-	value = value_from_property(match, "list-style-type");
+	value = value_from_property(match, PRO_LIST_STYLE_TYPE);
 	if (value)
 	{
 		if (!strcmp(value->data, "none")) style->list_style_type = LST_NONE;
@@ -1259,68 +1291,252 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		else if (!strcmp(value->data, "georgian")) style->list_style_type = LST_GEORGIAN;
 	}
 
-	style->line_height = number_from_property(match, "line-height", 1.2f, N_SCALE);
+	style->line_height = number_from_property(match, PRO_LINE_HEIGHT, 1.2f, N_SCALE);
 
-	style->text_indent = number_from_property(match, "text-indent", 0, N_LENGTH);
+	style->text_indent = number_from_property(match, PRO_TEXT_INDENT, 0, N_LENGTH);
 
-	style->width = number_from_property(match, "width", 0, N_AUTO);
-	style->height = number_from_property(match, "height", 0, N_AUTO);
+	style->width = number_from_property(match, PRO_WIDTH, 0, N_AUTO);
+	style->height = number_from_property(match, PRO_HEIGHT, 0, N_AUTO);
 
-	style->margin[0] = number_from_property(match, "margin-top", 0, N_LENGTH);
-	style->margin[1] = number_from_property(match, "margin-right", 0, N_LENGTH);
-	style->margin[2] = number_from_property(match, "margin-bottom", 0, N_LENGTH);
-	style->margin[3] = number_from_property(match, "margin-left", 0, N_LENGTH);
+	style->margin[0] = number_from_property(match, PRO_MARGIN_TOP, 0, N_LENGTH);
+	style->margin[1] = number_from_property(match, PRO_MARGIN_RIGHT, 0, N_LENGTH);
+	style->margin[2] = number_from_property(match, PRO_MARGIN_BOTTOM, 0, N_LENGTH);
+	style->margin[3] = number_from_property(match, PRO_MARGIN_LEFT, 0, N_LENGTH);
 
-	style->padding[0] = number_from_property(match, "padding-top", 0, N_LENGTH);
-	style->padding[1] = number_from_property(match, "padding-right", 0, N_LENGTH);
-	style->padding[2] = number_from_property(match, "padding-bottom", 0, N_LENGTH);
-	style->padding[3] = number_from_property(match, "padding-left", 0, N_LENGTH);
+	style->padding[0] = number_from_property(match, PRO_PADDING_TOP, 0, N_LENGTH);
+	style->padding[1] = number_from_property(match, PRO_PADDING_RIGHT, 0, N_LENGTH);
+	style->padding[2] = number_from_property(match, PRO_PADDING_BOTTOM, 0, N_LENGTH);
+	style->padding[3] = number_from_property(match, PRO_PADDING_LEFT, 0, N_LENGTH);
 
-	style->color = color_from_property(match, "color", black);
-	style->background_color = color_from_property(match, "background-color", transparent);
+	style->color = color_from_property(match, PRO_COLOR, black);
+	style->background_color = color_from_property(match, PRO_BACKGROUND_COLOR, transparent);
 
-	style->border_style_0 = border_style_from_property(match, "border-top-style");
-	style->border_style_1 = border_style_from_property(match, "border-right-style");
-	style->border_style_2 = border_style_from_property(match, "border-bottom-style");
-	style->border_style_3 = border_style_from_property(match, "border-left-style");
+	style->border_style_0 = border_style_from_property(match, PRO_BORDER_TOP_STYLE);
+	style->border_style_1 = border_style_from_property(match, PRO_BORDER_RIGHT_STYLE);
+	style->border_style_2 = border_style_from_property(match, PRO_BORDER_BOTTOM_STYLE);
+	style->border_style_3 = border_style_from_property(match, PRO_BORDER_LEFT_STYLE);
 
-	style->border_color[0] = color_from_property(match, "border-top-color", style->color);
-	style->border_color[1] = color_from_property(match, "border-right-color", style->color);
-	style->border_color[2] = color_from_property(match, "border-bottom-color", style->color);
-	style->border_color[3] = color_from_property(match, "border-left-color", style->color);
+	style->border_color[0] = color_from_property(match, PRO_BORDER_TOP_COLOR, style->color);
+	style->border_color[1] = color_from_property(match, PRO_BORDER_RIGHT_COLOR, style->color);
+	style->border_color[2] = color_from_property(match, PRO_BORDER_BOTTOM_COLOR, style->color);
+	style->border_color[3] = color_from_property(match, PRO_BORDER_LEFT_COLOR, style->color);
 
-	style->border_width[0] = border_width_from_property(match, "border-top-width");
-	style->border_width[1] = border_width_from_property(match, "border-right-width");
-	style->border_width[2] = border_width_from_property(match, "border-bottom-width");
-	style->border_width[3] = border_width_from_property(match, "border-left-width");
+	style->border_width[0] = border_width_from_property(match, PRO_BORDER_TOP_WIDTH);
+	style->border_width[1] = border_width_from_property(match, PRO_BORDER_RIGHT_WIDTH);
+	style->border_width[2] = border_width_from_property(match, PRO_BORDER_BOTTOM_WIDTH);
+	style->border_width[3] = border_width_from_property(match, PRO_BORDER_LEFT_WIDTH);
 
 	{
-		const char *font_weight = string_from_property(match, "font-weight", "normal");
-		const char *font_style = string_from_property(match, "font-style", "normal");
+		const char *font_weight = string_from_property(match, PRO_FONT_WEIGHT, "normal");
+		const char *font_style = string_from_property(match, PRO_FONT_STYLE, "normal");
+		const char *font_variant = string_from_property(match, PRO_FONT_VARIANT, "normal");
 		int is_bold = is_bold_from_font_weight(font_weight);
 		int is_italic = is_italic_from_font_style(font_style);
-		value = value_from_property(match, "font-family");
+		style->small_caps = !strcmp(font_variant, "small-caps");
+		value = value_from_property(match, PRO_FONT_FAMILY);
 		while (value)
 		{
 			if (strcmp(value->data, ",") != 0)
 			{
-				style->font = fz_load_html_font(ctx, set, value->data, is_bold, is_italic);
+				style->font = fz_load_html_font(ctx, set, value->data, is_bold, is_italic, style->small_caps);
 				if (style->font)
 					break;
 			}
 			value = value->next;
 		}
 		if (!style->font)
-			style->font = fz_load_html_font(ctx, set, "serif", is_bold, is_italic);
+			style->font = fz_load_html_font(ctx, set, "serif", is_bold, is_italic, style->small_caps);
 	}
 }
 
+#ifdef DEBUG_CSS_SPLAY
+static void
+do_verify_splay(const fz_css_style_splay *x)
+{
+	printf("%x<", x);
+	if (x->lt)
+	{
+		assert(memcmp(&x->lt->style, &x->style, sizeof(x->style)) < 0);
+		assert(x->lt->up == x);
+		do_verify_splay(x->lt);
+	}
+	printf(",");
+	if (x->gt)
+	{
+		assert(memcmp(&x->gt->style, &x->style, sizeof(x->style)) > 0);
+		assert(x->gt->up == x);
+		do_verify_splay(x->gt);
+	}
+	printf(">\n");
+}
+
+static void
+verify_splay(const fz_css_style_splay *x)
+{
+	if (x == NULL)
+		return;
+	assert(x->up == NULL);
+	do_verify_splay(x);
+	printf("-----\n");
+}
+#endif
+
+const fz_css_style *
+fz_css_enlist(fz_context *ctx, const fz_css_style *style, fz_css_style_splay **tree, fz_pool *pool)
+{
+	fz_css_style_splay **current = tree;
+	fz_css_style_splay *x;
+	fz_css_style_splay *y = NULL;
+
+	/* Search for a match in the tree, if there is one, or for
+	 * the insertion point, if there is not. */
+	while (*current != NULL)
+	{
+		int cmp = memcmp(style, &(*current)->style, sizeof(*style));
+		if (cmp == 0)
+		{
+			/* We have a match - break out and do move to root. */
+			break;
+		}
+		y = (*current);
+		if (cmp < 0)
+			current = &y->lt;
+		else
+			current = &y->gt;
+	}
+	/* Create one if needed */
+	if (*current == NULL)
+	{
+		x = *current = fz_pool_alloc(ctx, pool, sizeof(*y));
+		x->style = *style;
+		x->up = y;
+		x->lt = NULL;
+		x->gt = NULL;
+	}
+	else
+		x = *current;
+	/* Now move to root */
+	/*
+	The splaying steps used:
+
+	Case 1:	|a)       z              x             b)     z                   x
+		|     y       D  =>  A       y            A       y           y       D
+		|   x   C                  B   z                B   x  =>  z     C
+		|  A B                        C D                  C D    A B
+
+	Case 2:	|a)       z              x             b)     z                   x
+		|     y       D  =>   y     z             A       y    =>     z       y
+		|   A   x            A B   C D                  x   D        A B     C D
+		|      B C                                     B C
+
+	Case 3:	|a)       y              x             b)     y                   x
+		|      x     C   =>   A     y              A     x      =>     y     C
+		|     A B                  B C                  B C           A B
+	*/
+#ifdef DEBUG_CSS_SPLAY
+	printf("BEFORE\n");
+	verify_splay(*tree);
+#endif
+	while ((y = x->up) != NULL ) /* While we're not at the root */
+	{
+		fz_css_style_splay *z = y->up;
+		y->up = x;
+		if (z == NULL)
+		{
+			if (y->lt == x)	/* Case 3a */
+			{
+				y->lt = x->gt;
+				if (y->lt)
+					y->lt->up = y;
+				x->gt = y;
+			}
+			else /* Case 3b */
+			{
+				y->gt = x->lt;
+				if (y->gt)
+					y->gt->up = y;
+				x->lt = y;
+			}
+			x->up = NULL;
+			break;
+		}
+		x->up = z->up;
+		if (z->up)
+		{
+			if (z->up->lt == z)
+				z->up->lt = x;
+			else
+				z->up->gt = x;
+		}
+		if (z->lt == y)
+		{
+			if (y->lt == x) /* Case 1a */
+			{
+				z->lt = y->gt;
+				if (z->lt)
+					z->lt->up = z;
+				y->lt = x->gt;
+				if (y->lt)
+					y->lt->up = y;
+				y->gt = z;
+				z->up = y;
+				x->gt = y;
+			}
+			else /* Case 2a */
+			{
+				y->gt = x->lt;
+				if (y->gt)
+					y->gt->up = y;
+				z->lt = x->gt;
+				if (z->lt)
+					z->lt->up = z;
+				x->lt = y;
+				x->gt = z;
+				z->up = x;
+			}
+		}
+		else
+		{
+			if (y->gt == x) /* Case 1b */
+			{
+				z->gt = y->lt;
+				if (z->gt)
+					z->gt->up = z;
+				y->gt = x->lt;
+				if (y->gt)
+					y->gt->up = y;
+				y->lt = z;
+				z->up = y;
+				x->lt = y;
+			}
+			else /* Case 2b */
+			{
+				z->gt = x->lt;
+				if (z->gt)
+					z->gt->up = z;
+				y->lt = x->gt;
+				if (y->lt)
+					y->lt->up = y;
+				x->gt = y;
+				x->lt = z;
+				z->up = x;
+			}
+		}
+	}
+
+	*tree = x;
+#ifdef DEBUG_CSS_SPLAY
+	printf("AFTER\n");
+	verify_splay(x);
+#endif
+
+	return &x->style;
+}
 /*
  * Pretty printing
  */
 
-void
-print_value(fz_css_value *val)
+static void print_value(fz_css_value *val)
 {
 	printf("%s", val->data);
 	if (val->args)
@@ -1336,18 +1552,16 @@ print_value(fz_css_value *val)
 	}
 }
 
-void
-print_property(fz_css_property *prop)
+static void print_property(fz_css_property *prop)
 {
-	printf("\t%s: ", prop->name);
+	printf("\t%s: ", fz_css_property_name(prop->name));
 	print_value(prop->value);
 	if (prop->important)
 		printf(" !important");
 	printf(";\n");
 }
 
-void
-print_condition(fz_css_condition *cond)
+static void print_condition(fz_css_condition *cond)
 {
 	if (cond->type == '=')
 		printf("[%s=%s]", cond->key, cond->val);
@@ -1359,19 +1573,16 @@ print_condition(fz_css_condition *cond)
 		print_condition(cond->next);
 }
 
-void
-print_selector(fz_css_selector *sel)
+static void print_selector(fz_css_selector *sel)
 {
 	if (sel->combine)
 	{
-putchar('(');
 		print_selector(sel->left);
 		if (sel->combine == ' ')
 			printf(" ");
 		else
 			printf(" %c ", sel->combine);
 		print_selector(sel->right);
-putchar(')');
 	}
 	else if (sel->name)
 		printf("%s", sel->name);
@@ -1383,8 +1594,7 @@ putchar(')');
 	}
 }
 
-void
-print_rule(fz_css_rule *rule)
+static void print_rule(fz_css_rule *rule)
 {
 	fz_css_selector *sel;
 	fz_css_property *prop;
@@ -1406,8 +1616,9 @@ print_rule(fz_css_rule *rule)
 }
 
 void
-print_rules(fz_css_rule *rule)
+fz_debug_css(fz_context *ctx, fz_css *css)
 {
+	fz_css_rule *rule = css->rule;
 	while (rule)
 	{
 		print_rule(rule);
